module HarmTrace.Audio.Utils where

import HarmTrace.Audio.ChordTypes

import Data.List (genericLength, transpose, tails, inits,maximumBy, sort, group)
import Data.Ord (comparing)

--------------------------------------------------------------------------------
-- Some utility functions used by other functions throughout the Audio module
--------------------------------------------------------------------------------

-- returns the last timestamp
-- getLastFrameTimeStamp :: [ChordinoLine] -> NumData
-- getLastFrameTimeStamp = time . last

-- given a list of beats, a list of grouped data items, and a merging function
-- mergeAndTimeStamp returns a list of TimedData items of which these
-- data items are grouped with the provided merging function
mergeAndTimeStamp  ::([a] -> a)-> [NumData] -> [[a]] -> [TimedData a]
mergeAndTimeStamp f = mergeAndTimeStamp' f 0
mergeAndTimeStamp' ::([a] -> a)-> NumData -> [NumData] -> [[a]] -> [TimedData a]
mergeAndTimeStamp' _      _   []    []       = []
mergeAndTimeStamp' mergef ons beats (x : xs) = 
  let (offs : rest) = drop ((length x) -1) beats
  in  TimedData (mergef x) ons offs : mergeAndTimeStamp' mergef offs rest xs
mergeAndTimeStamp' _ _ _ _ = 
  error "mergeAndTimeStamp: asynchronous beats and data"
  
-- a function for merging identical lists of ProbChords. Equallity of the chords
-- in the lists is assumed and the probabilities are averaged
avgPC :: [[ProbChord]] -> [ProbChord]
avgPC [[a]] = [a]
avgPC pcs   = let cs = map chordLab $ head pcs 
                  ps = map listMean . transpose $ map (map prob) pcs
              in  zipWith ProbChord cs ps  

-- similarly to group, but in case a group is smaller than s the group is filled
-- with the same number of a's
-- ?> groupMinSize 3 0 [1,2,2,3,3,3,4,4,4,4,5,5,5,5,5]
-- [[0],[0,0],[0,0,0],[4,4,4,4],[5,5,5,5,5]]
groupMinSize :: Eq a => Int -> a -> [a] -> [[a]]
groupMinSize _ _  []    =  []
groupMinSize s a (x:xs) =  grp : groupMinSize s a zs
   where (ys,zs) = span (== x) xs
         lys = length ys
         grp = if length ys >= s then x:ys else replicate (lys + 1) a
         
-- samples at specific points in time, specified in a list
sampleAt :: [NumData] -> [TimedData a] -> [a]
sampleAt  _     [] = []
sampleAt []     a  = map getData a -- will never occur, return without sampling
sampleAt (t:ts) (c:cs)
  | t <= offset c = getData c : sampleAt ts (c:cs)
  | otherwise     = sampleAt (t:ts) cs         

--------------------------------------------------------------------------------
-- Statistical functions
--------------------------------------------------------------------------------
              
listMean :: [NumData] -> NumData
listMean a = sum a  / genericLength a
  
-- a median filter: see http://en.wikipedia.org/wiki/Median_filter
medianFilter :: Ord a => Int -> [a] -> [a]
medianFilter wsize l = map mode $ getWindows wsize l
  
-- returns a list with all "sliding windows" of size wsize  
-- the left and right edge of the list are with the first and last (size /2)
-- items, respectively and the remainder is filled with the mode/median of
-- the complete list
getWindows :: Ord a => Int -> [a] -> [[a]]
getWindows size l = lbor ++ mid ++ rbor
  where mid  = takeWhile (hasSize size) . map (take size) $ tails l
        gmed = mode l
        ls   = size `div` 2
        rs   = size - ls - 1
        lbor = reverse $ map (fillWith size gmed) 
                             (reverse . dropWhile (null) . inits $ take ls l)
        rbor = map (reverse . fillWith size gmed) 
                   (takeWhile (not . null) . tails . take rs $ reverse l)
        -- in general (length l) < x, but this is guaranteed within this let
        fillWith ::  Int -> a -> [a] -> [a]
        fillWith x a lt = replicate (x - (length l)) a ++ lt

hasSize :: Int -> [a] -> Bool
hasSize s l = length l >= s
  
-- the mode: the element that occurs most often in the collection 
mode   :: Ord a => [a] -> a
mode [] = error "Key.hs: mode called on []"
mode l  = head . maximumBy (comparing length) . group $ sort l 
  
-- the median: not good for keys... 
median   :: Ord a => [a] -> a
median [] = error "Key.hs: median called on []"
median l  = sort l !! (length l `div` 2)  

maxListIndex :: Ord a => [a] -> Int
maxListIndex = fst . maxPair where
  maxPair :: Ord a => [a] -> (Int,a)
  maxPair = maximumBy (comparing snd) . zip [0..] 

--------------------------------------------------------------------------------
-- Making beat synchronized chroma
--------------------------------------------------------------------------------
mergeBeats :: Int -> [[a]] -> [[a]]
mergeBeats _ [] = []
mergeBeats n l  = let (a,b) = splitAt n l in concat a : mergeBeats n b

-- synchronizes the ChordinoData to the list of beats in the BeatTrackerData
-- by grouping the ChordinoLines of the ChordinoData in separate lists
beatSync :: BeatTrackerData -> ChordinoData -> BeatChroma
beatSync _   []     = []
beatSync bt (cd:cs) = beatSync' bt [cd] (cd:cs) where
  beatSync' :: [NumData] -> [ChordinoLine] -> [ChordinoLine] -> [[ChordinoLine]]
  beatSync' _  _   [] = []
  beatSync' [] _   c  = [c]
  beatSync' (b:bs) prv c -- we also store the previous group in case beat < time
    | null x    = prv : beatSync' bs prv xs 
    | otherwise = x   : beatSync' bs x   xs 
        where (x, xs) = span ((>=) b . time) c